// Copyright (C) 2009 Playedit.

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "config.h"

#include "music.h"

#include <glibmm/i18n.h>
#include <gtkmm/stock.h>
#include <iomanip>

MusicView::MusicView(Playedit::App& app_window)
	: DNDTreeViewBase(_("Music")),
	m_App(app_window)
{
	m_Column.pack_start(m_AlbumArt, false);
	m_Column.set_cell_data_func(m_AlbumArt, sigc::mem_fun(*this, &MusicView::AlbumArtDataFunc));

	m_Column.pack_start(m_MusicName);
	m_Column.set_cell_data_func(m_MusicName, sigc::mem_fun(*this, &MusicView::MusicNameDataFunc));

	// Set up row selections
	get_selection()->set_select_function(sigc::mem_fun(*this, &MusicView::SelectionFunc));
	get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);

	// Set up sorting
	m_Store->set_default_sort_func(sigc::mem_fun(*this, &MusicView::SortByName));
	m_Store->set_sort_column(Gtk::TreeSortable::DEFAULT_SORT_COLUMN_ID, Gtk::SORT_ASCENDING);
	m_Column.set_clickable();
	m_Column.signal_clicked().connect(sigc::mem_fun(*this, &MusicView::SelectSortTypePopup));

	// Setup as a drag destination
	std::vector<Gtk::TargetEntry> target_entries;
	target_entries.push_back(Gtk::TargetEntry("text/uri-list", Gtk::TARGET_OTHER_APP, 0));
	drag_dest_set(target_entries, Gtk::DEST_DEFAULT_ALL, Gdk::ACTION_COPY);

	// Setup row changed signal to set visibility of the appropriate actions
	// and do it for the first time.
	get_selection()->signal_changed().connect(sigc::mem_fun(*this, &MusicView::SelectionChanged));
	SelectionChanged();
}

// Scale the album art down to a reasonable size for the treeview.
void MusicView::AlbumArtDataFunc(Gtk::CellRenderer *renderer,
		const Gtk::TreeIter& iter)
{
	if (Glib::RefPtr<album_t> album =
			Glib::RefPtr<album_t>::cast_dynamic(iter->get_value(m_StoreColumns.m_Data)))
	{
		if (album->art)
		{
			((Gtk::CellRendererPixbuf*)renderer)->property_stock_id().reset_value();
			Glib::RefPtr<Gdk::Pixbuf> scaled_album_art =
				album->art->scale_simple(24, 24, Gdk::INTERP_BILINEAR);
			((Gtk::CellRendererPixbuf*)renderer)->property_pixbuf() = scaled_album_art;
		}
		else
			((Gtk::CellRendererPixbuf*)renderer)->property_stock_id() =
				Gtk::Stock::MISSING_IMAGE.id;
		renderer->set_visible();
	}
	else
		renderer->set_visible(false);
}

// Set the text that is shown in the music treeview.
void MusicView::MusicNameDataFunc(Gtk::CellRenderer *renderer,
		const Gtk::TreeIter& iter)
{
	Glib::RefPtr<mtp_data_t> mtp_data = (*iter)[m_StoreColumns.m_Data];
	if (Glib::RefPtr<track_t> track = Glib::RefPtr<track_t>::cast_dynamic(mtp_data))
	{
		if (track->TrackNumber())
			((Gtk::CellRendererText*)renderer)->property_text() =
				Glib::ustring::format(std::setw(2), track->TrackNumber(), ". ", track->Name());
		else
			((Gtk::CellRendererText*)renderer)->property_text() = mtp_data->Name();
	}
	else
		((Gtk::CellRendererText*)renderer)->property_text() = mtp_data->Name();
}

// Add the albums and tracks to the music store so they will be displayed.
void MusicView::PopulateStore(const std::vector<album_pair_t>& album_list)
{
	for (std::vector<album_pair_t>::const_iterator album = album_list.begin();
			album != album_list.end(); ++album)
	{
		Gtk::TreeRow album_row = *(m_Store->append());
		album_row[m_StoreColumns.m_Data] = album->first;

		for (std::vector< Glib::RefPtr<track_t> >::const_iterator album_track = album->second.begin();
				album_track != album->second.end(); ++album_track)
		{
			Gtk::TreeRow track_row = *(m_Store->append(album_row.children()));
			track_row[m_StoreColumns.m_Data] = *album_track;
		}
	}
}

// Only allow rows at the same level to be in a multiple selection.
bool MusicView::SelectionFunc(const Glib::RefPtr<Gtk::TreeModel>& /* model */,
		const Gtk::TreePath& path, bool /* path_currently_selected */)
{
	std::vector<Gtk::TreePath> selected_rows = get_selection()->get_selected_rows();
	if (selected_rows.size() && selected_rows.front().size() != path.size())
		return false;

	return true;
}

void MusicView::SetSortByName()
{
	m_Store->set_default_sort_func(sigc::mem_fun(*this, &MusicView::SortByName));
}

void MusicView::SetSortByArtist()
{
	m_Store->set_default_sort_func(sigc::mem_fun(*this, &MusicView::SortByArtist));
}

void MusicView::SelectSortTypePopup()
{
	Gtk::Menu* music_sort_popup =
		(Gtk::Menu*)m_App.m_UIManager->get_widget("/MusicSortPopup");
	music_sort_popup->popup(1, GDK_CURRENT_TIME);
}

int MusicView::SortByName(const Gtk::TreeIter& a, const Gtk::TreeIter& b)
{
	Glib::RefPtr<album_t> album_a =
		Glib::RefPtr<album_t>::cast_dynamic(a->get_value(m_StoreColumns.m_Data));
	Glib::RefPtr<album_t> album_b =
		Glib::RefPtr<album_t>::cast_dynamic(b->get_value(m_StoreColumns.m_Data));
	if (album_a && album_b)
		return AlbumCompareNameFirst(album_a, album_b);

	Glib::RefPtr<track_t> track_a =
		Glib::RefPtr<track_t>::cast_dynamic(a->get_value(m_StoreColumns.m_Data));
	Glib::RefPtr<track_t> track_b =
		Glib::RefPtr<track_t>::cast_dynamic(b->get_value(m_StoreColumns.m_Data));
	if (track_a && track_b)
		return TrackCompare(track_a, track_b);

	return 0;
}

int MusicView::SortByArtist(const Gtk::TreeIter& a, const Gtk::TreeIter& b)
{
	Glib::RefPtr<album_t> album_a =
		Glib::RefPtr<album_t>::cast_dynamic(a->get_value(m_StoreColumns.m_Data));
	Glib::RefPtr<album_t> album_b =
		Glib::RefPtr<album_t>::cast_dynamic(b->get_value(m_StoreColumns.m_Data));
	if (album_a && album_b)
		return AlbumCompareArtistFirst(album_a, album_b);

	Glib::RefPtr<track_t> track_a =
		Glib::RefPtr<track_t>::cast_dynamic(a->get_value(m_StoreColumns.m_Data));
	Glib::RefPtr<track_t> track_b =
		Glib::RefPtr<track_t>::cast_dynamic(b->get_value(m_StoreColumns.m_Data));
	if (track_a && track_b)
		return TrackCompare(track_a, track_b);

	return 0;
}

bool MusicView::on_button_press_event(GdkEventButton* event)
{
	Gtk::TreePath path;
	if (m_App.m_Device.IsReady() && event->type == GDK_BUTTON_PRESS &&
			get_path_at_pos(event->x, event->y, path))
	{
		if (event->button == 3)
		{
			if (!get_selection()->is_selected(path))
			{
				get_selection()->unselect_all();
				get_selection()->select(path);
			}

			Gtk::Menu* music_popup =
				(Gtk::Menu*)m_App.m_UIManager->get_widget("/MusicPopup");
			music_popup->popup(event->button, event->time);
			return true;
		}

		if (event->button == 1)
		{
			if (!(event->state & (Gdk::SHIFT_MASK | Gdk::CONTROL_MASK))
					&& (event->x > EXPANDER_WIDTH))
			{
				m_DragX = event->x;
				m_DragY = event->y;

				if (get_selection()->is_selected(path) &&
						get_selection()->count_selected_rows() != 1)
					return true;
			}
			// Don't allow shift selection to leave the current level
			else if (event->state & Gdk::SHIFT_MASK)
			{
				std::vector<Gtk::TreePath> selected_rows =
					get_selection()->get_selected_rows();
				if (selected_rows.size() &&
						selected_rows.front().size() != path.size())
					return true;
			}
		}
	}

	return Gtk::TreeView::on_button_press_event(event);
}

// Set the sensitivity of the items in the music popup appropriately.
void MusicView::SelectionChanged()
{
	int num_selected = get_selection()->count_selected_rows();
	int depth = num_selected ?
		std::vector<Gtk::TreePath>(get_selection()->get_selected_rows()).front().size() :
		0;

	m_App.m_UIManager->get_action("/MusicPopup/EditAlbumMetadata")->
		set_sensitive(num_selected == 1 && depth == 1);
	m_App.m_UIManager->get_action("/MusicPopup/EditAlbumMetadata")->
		set_visible(num_selected == 1 && depth == 1);
	m_App.m_UIManager->get_action("/MusicPopup/AddAlbumArt")->
		set_sensitive(num_selected == 1 && depth == 1);
	m_App.m_UIManager->get_action("/MusicPopup/AddAlbumArt")->
		set_visible(num_selected == 1 && depth == 1);
	m_App.m_UIManager->get_action("/MusicPopup/EditTrackMetadata")->
		set_sensitive(num_selected == 1 && depth == 2);
	m_App.m_UIManager->get_action("/MusicPopup/EditTrackMetadata")->
		set_visible(num_selected == 1 && depth == 2);
	m_App.m_UIManager->get_action("/MusicPopup/DeleteTracks")->
		set_sensitive(num_selected && depth == 2);
	m_App.m_UIManager->get_action("/MusicPopup/DeleteTracks")->
		set_visible(num_selected && depth == 2);
	m_App.m_UIManager->get_action("/MusicPopup/DeleteAlbums")->
		set_sensitive(num_selected && depth == 1);
	m_App.m_UIManager->get_action("/MusicPopup/DeleteAlbums")->
		set_visible(num_selected && depth == 1);
}

// Set the drag icon if drug from the music treeview
void MusicView::on_drag_begin(const Glib::RefPtr<Gdk::DragContext>& context)
{
	Glib::RefPtr<Gtk::TreeModel> model;
	std::vector<Gtk::TreePath> selected_cells = get_selection()->get_selected_rows(model);

	int x = 0, y = 0;
	std::vector< Glib::RefPtr<Pango::Layout> > drag_icon_rows;
	for (std::vector<Gtk::TreePath>::iterator selected = selected_cells.begin();
			selected != selected_cells.end(); ++selected)
	{
		Glib::RefPtr<mtp_data_t> mtp_data = (*model->get_iter(*selected))[m_StoreColumns.m_Data];
		drag_icon_rows.push_back(create_pango_layout(mtp_data->Name()));

		int width, height;
		drag_icon_rows.back()->get_size(width, height);
		x = std::max(x, (width / PANGO_SCALE));
		y = std::max(y, (height / PANGO_SCALE));
	}

	Cairo::RefPtr<Cairo::Surface> surface =
		get_bin_window()->create_similar_surface(Cairo::CONTENT_COLOR,
				x + 30, (y + 6) * selected_cells.size() + 2);

    Cairo::RefPtr<Cairo::Context> cairo_ctx = Cairo::Context::create(surface);

	// Fill the rectangle with a white background
	cairo_ctx->set_source_rgba(1.0, 1.0, 1.0, 1.0); // white
	cairo_ctx->rectangle(1, 1, x + 28, (y + 6) * selected_cells.size());
	cairo_ctx->fill();

	// Draw the text in black
	cairo_ctx->set_source_rgba(0.0, 0.0, 0.0, 1.0); // black
	int count = 0;
	for (std::vector< Glib::RefPtr<Pango::Layout> >::iterator row = drag_icon_rows.begin();
			row != drag_icon_rows.end(); ++row, ++count)
	{
		cairo_ctx->move_to(15, 3 + count * (y + 6));
		(*row)->show_in_cairo_context(cairo_ctx);

		// Draw a gray line between the text lines
		if (count > 0)
		{
			cairo_ctx->save();
			cairo_ctx->set_source_rgba(0.7, 0.7, 0.7, 1.0); // gray
			cairo_ctx->set_line_width(1.0);
			cairo_ctx->move_to(1, count * (y + 6) + 0.5);
			cairo_ctx->line_to(x + 29, count * (y + 6) + 0.5);
			cairo_ctx->stroke();
			cairo_ctx->restore();
		}
	}

	context->set_icon(cairo_ctx->get_target());
}
